/* This is an implementation of shift scheduling problem
 * by S. Moriguchi and N. Tsuchimura.
 * This program is a part of demonstration CGI.
 * It is called from 'demo-shift.cgi' via 'odicon.cgi'.
 *
 *  c.f. G. Koole and E. van der Sluis (2003): Optimal shift scheduling
 *  with a global service level constraint, IIE Transactions, Vol. 35,
 *  pp. 1049-1055.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "odicon.h"
#include "con_util.h"

#define INTERVAL 13
#define DIM 3

/* determined in demo_shift_schedule() */
static double c = 0.01, mu = 20.0;
static double lambda[INTERVAL] = {88, 87, 86, 85, 55, 1, 1, 1, 1, 1, 1, 1, 1};
static int n_total;
static int salary[DIM+1];/* = {2, 1, 1, 1}; */
static double Gr = 100;
/*static int Ds[DIM+1];*/   /* needless Ds*/

/* needless Ds */
/*
static void D_inv(int x[]){
    int i, sum=0;

    for (i=DIM; i>=0; i--) {
	sum += x[i];
    	salary[i] = sum;
    }
}
*/

static double si(int n, int i) {
    return salary[i-1]*n;
}

/* x^n / n! */
static double exp_per_fact(double x, int n) {
    int i;
    double r = 1.0;

    for (i=n; i>=1; i--) r = r * x / i;
    return r;
}

/* pi = b / (a + b), which is safe even if b=inf */
static double pi(double a, double b) {
    if (b < a) {
        double s = b / a;
	return s / ( 1 + s );
    } else {
        double t = a / b;
	return 1 / ( t + 1 );
    }
}


static double gi(int n, int i, int *outside) {
    double rho, rho_n, alpha, beta, gi;
    int j;
    
    rho = lambda[i-1] / mu;
    rho_n = rho / n;
 
    if (rho_n >=1) {
        *outside = TRUE;
        return PLUS_INFINITY;
    }
    beta = exp_per_fact(rho,n) / (1-rho_n);
    
    alpha = 0;
    for (j=0; j<n; j++) {
        alpha += exp_per_fact(rho,j);
    }

    gi = lambda[i-1] * pi(alpha, beta) * exp(-n * mu*(1-rho_n)*c);

    /*
    printf(" gi[%d] = %g\trho=%g\ta=%g\tb=%g\tpi=%g\texp(*)=%g\n",
	   i, gi, rho, alpha, beta, pi(alpha, beta), exp(-n * mu*(1-rho_n)*c));
    */
    return gi;
}

/* variable translation from L#-convex to multimodular */
static void x2y(int dim, int x[], int y[]) {
    int i, last = 0;

    for (i=0; i<dim; i++) {
        y[i] = x[i] - last;
        last = x[i];
    }
}

/* variable translation from multimodular to L#-convex */
static void y2x(int dim, int y[], int x[]) {
    int i, sum = 0;

    for (i=0; i<dim; i++) {
        sum += y[i];
        x[i] = sum;
    }
}


/* f1(x) = \sum g_i */
static double f1(int dim, int x[]) {
    double t, LAMBDA, s;
    int outside = FALSE;
    int j;
  
    assert(dim == 3);  /* check dimension */

    if (x[0] < 0) return PLUS_INFINITY;
    for (j=1; j<dim; j++) {
        if (x[j] < 0 || x[j]<x[j-1]) return PLUS_INFINITY;
    }
    for (j=0; j<dim; j++) {
        assert(lambda[j] < n_total * mu);
    }

    t = 0;
    t +=  gi(x[0]-0, 1, &outside);
    t +=  gi(x[0]-0, 2, &outside);
    t +=  gi(x[1]-0, 3, &outside);
    t +=  gi(x[1]-0, 4, &outside);
    t +=  gi(x[1]-0, 5, &outside);
    t +=  gi(x[2]-x[0], 6, &outside);
    t +=  gi(x[2]-x[0], 7, &outside);
    t +=  gi(x[2]-x[1], 8, &outside);
    t +=  gi(n_total-x[1], 9, &outside);
    t +=  gi(n_total-x[1], 10, &outside);
    t +=  gi(n_total-x[2], 11, &outside);
    t +=  gi(n_total-x[2], 12, &outside);
    t +=  gi(n_total-x[2], 13, &outside);

    LAMBDA = 0;
    for (j=0; j<INTERVAL; j++) {
        LAMBDA += lambda[j];
    }
    /*printf(" LAMBDA = %g\n\n", LAMBDA);*/

    s = 0;
    s +=  si(x[0]-0, 1);
    s +=  si(x[1]-x[0], 2);
    s +=  si(x[2]-x[1], 3);
    s +=  si(n_total-x[2], 4);

    if (outside) return PLUS_INFINITY;
    return Gr * t / LAMBDA + s;
}

static double conjugate(int dim, int y[]) {
    static int x[DIM+1] = { 5,5,11,16 };
    double ret;
    enum verbose_e vl;

    y2x(dim+1, y, salary);
    /*printf("salary = "); ivector_print(dim+1, salary); printf("\n");*/
    n_total = x[DIM];

    vl = verbose_level_get();
    verbose_level_set(0);
    ret = lgconv_minimize(dim, f1, x);
    verbose_level_set(vl);
    /*printf("x = "); ivector_print(dim+1, x); printf("\n");*/
    return ret;
}

void demo_shift_schedule2(const int dim) {
    int y[DIM+1] = {1, 1, 1, 7};  /* initial solution, multimodular */
    int x[DIM+1];                 /* L# variable */
    int lower[DIM+1] = { 1, 1, 1, 1 };
    int upper[DIM+1] = { 10, 10, 10, 10 };
    int sum = 20; /* for M-convex */
    double ans;
    int i;

    assert( dim == DIM );
    /* printf("gi = %f\n", gi(5,5)); */

    fvector_scan(1, &c);
    fvector_scan(1, &mu);
    fvector_scan(1, &Gr);
    fvector_scan(INTERVAL, lambda);
    ivector_scan(DIM+1, y);
    /*ivector_scan(DIM+1, Ds);*/ /* needless Ds*/
    ivector_scan(DIM+1, salary);

    y2x(DIM+1, y, x);
    n_total = x[DIM];
    /*D_inv(Ds);*/ /* needless Ds*/
    ans = lgconv_minimize(DIM, f1, x);

    /*printf("S(x) = %f\n", -ans);*/
    printf("\nN = %d\n", n_total);
    x2y(DIM+1, x, y);
    printf("\n");
    for (i=0; i<DIM+1; i++) {
        printf("x[%d] = %d\n", i+1, x[i]);
    }
    printf("\n");
    for (i=0; i<DIM+1; i++) {
        printf("y[%d] = %d\n", i+1, y[i]);
    }

    printf("\nprimal ...\n");
    verbose_level_off(STATUS|COUNTER);
    RANGE_PRINT(DIM);
    CON_PRINT(DIM, f1);

    printf("conjugate ...\n");
    RANGE_PRINT(DIM);
    CON_PRINT(DIM, conjugate);
}
