#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "serialfunc.h"

/* fserialfunc = sum(f(x)) */
/*   f(x) = ax^4 + bx^3 + cx^2 + dx + e  (type '4') */
/*   f(x) = ax^2 + bx + c                (type '2') */
/*   f(x) =  a * exp(bx + c)             (type 'E') */
/*   f(x) = -a * log(bx + c)             (type 'L') */
/*   f(x) =  a * abs(bx + c)             (type 'A') */

/* class L : laminar */
/* class P : Pseudo-separable */
static double fserialfunc(const int dim, double x[], double v[], char class) {
    double f = 0;
    int i=0, type;

    while ((type = (int)v[i++]) != '0') {  /* '0' is end mark */
        /* coefficient */
        double term = 0;
        double a=0, b=0, c=0, d=0, e=0;
        a = v[i++];
        b = v[i++];
        c = v[i++];
        switch (type) {
        case '2': case 'E': case 'L': case 'A':
            break;
        case '4':
            d = v[i++];
            e = v[i++];
        default:
            assert(0 != 0);  /* error */
            break;
        }

        /* term */
        switch (class) {
        case 'L':  /* laminar */
            while (v[i] >= 0) {
                assert(v[i] < dim);
                term += x[(int)v[i++]];
            }
            i++;
            break;
        case 'P': /* Pseudo-separable */
            assert(0 <= v[i] && v[i] < dim);
            term = x[(int)v[i++]];
            if (v[i] >= 0) {
                assert(v[i] < dim);
                term -= x[(int)v[i]];  /* not needed i++ */
            }
            i++;
            break;
        default:
            assert(0 != 0);  /* error */
            break;
        }

        switch (type) {
        case '4':
            f += a*term*term*term*term + b*term*term*term
                +c*term*term            +d*term           + e;
            break;
        case '2':
            f += a*term*term + b*term + c;
            break;
        case 'E':
            d = b*term + c;
            e = a * exp(d);
            if (e < DBL_MAX) {
                f += e;
            } else {
                fprintf(stderr, "exp(%g) is too big. Aborted.\n", d);
                exit(1);
            }
            break;
        case 'L':
            d = b*term + c;
            e = -a * log(d);
            if (e < DBL_MAX) {
                f += e;
            } else {
                fprintf(stderr, "log(%g) is not defined. Aborted.\n", d);
                exit(1);
            }
            break;
        case 'A':
            f += a * fabs(b*term + c);
            break;
        default:
            assert(0 != 0);  /* error */
            break;
        }
    }
    return f;
}


double flaminar(const int dim, double x[], double v[]) {
    return fserialfunc(dim, x, v, 'L');
}


double fpseparable(const int dim, double x[], double v[]) {
    return fserialfunc(dim, x, v, 'P');
}

double *fv;
char fclass;
static double f(const int dim, double x[]) {
    return fserialfunc(dim, x, fv, fclass);
}
double ilaminar(const int dim, int x[], double v[]) {
    fv = v;
    fclass = 'L';
    return int_func(dim, f, x);
}
double ipseparable(const int dim, int x[], double v[]) {
    fv = v;
    fclass = 'P';
    return int_func(dim, f, x);
}



/********** a sample program ***********/
#if TEST

#include <stdio.h>
#include <math.h>
#include "serialfunc.h"

void sample_for_flaminar(void) {
    double v[] = {
        'E', 2, 3, 4, /* 'E' needs three arguments */
        0, 1, 2, -1,  /* indexes of x[], -1 means end */
        '0',          /* end of equation */
    };
    double x[4] = { 9, 8, 7, 6 };
    double f = flaminar(4, x, v);
    
    printf("%f\n", f);
    printf("%f\n", 2*exp(3*(x[0]+x[1]+x[2])+4));
}


void sample_for_fpseparable(void) {
    double v[] = {
        'L', 2, 3, 4, /* 'L' needs three arguments */
        0, 1,         /* two indexes of x[], no end mark */
        '2', 5, 6, 7, /* '2' needs three arguments */
        3, -1,        /* one index of x[], -1 means ommited */
        '0',          /* end of equation */
    };
    double x[4] = { 9, 8, 7, 6 };
    double f = fpseparable(4, x, v);
    
    printf("%f\n", f);
    printf("%f\n", -2*log(3*(x[0]-x[1])+4) + 5*x[3]*x[3]+6*x[3]+7);
}

int main() {
    sample_for_flaminar();
    sample_for_fpseparable();
    return 0;
}
#endif /* TEST */
/********** a sample program ***********/
